#! /usr/bin/env python

# tdbsf_front_end.py --- outer interface for the TDBSF "search engine"
# Copyright (c) 2001, 2004 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""Front-ends to the Trivial Database Searching Facility.

Currently, this module holds a GNU Emacs interface and a dumb (for
demonstrating purposes) text interface.

"""

from __future__ import nested_scopes
import sys, os, os.path, string, getopt, types
import tdbsf_api
from tdbsf_exceptions import TDBSFGenericException


usage = """Usage: %s [OPTION]... EXPR PATTERN BOH EOH
Perform a TDBSF query.

EXPR                        search expression for the query
PATTERN                     file globbing pattern to use to determine the
                            set of files to examine
BOH                         Python regular expression specifying the
                            beginning of header of a record
EOH                         Python regular expression specifying the end
                            of header of a record (after which starts the
                            the body of the record)

Options:
      --boh-flags=FLAGS     flags to use for the regular expression search
                            performed when looking for the Beginning Of Header
      --eoh-flags=FLAGS     flags to use for the regular expression search
                            performed when looking for the End Of Header
      --crlf-to-lf-in-args  convert CRLF to LF in arguments before passing
                            them to the engine
      --dont-consider-eol-as-lf
                            don't try to guess the end-of-line type in
                            database files and convert it to line feed for use
                            by the engine; instead, tell the engine to work on
                            the raw file contents
  -m, --db-files-mode=MODE  mode to use when opening database files; permitted
                            values are "binary" ("b") and "text" ("t")
  -e, --emacs-friendly-output
                            generate output in a format that is easy to parse
                            in Emacs Lisp
      --header-body-separator=SEPARATOR
      --hbs=SEPARATOR       separator string to use between the header and the
                            body in the string where the search expression is
                            evaluated when the value given to --search-string
                            is header_plus_body
  -s, --search-string=WHERE string to use to evaluate the search expression;
                            permitted values are header, body,
                            header_plus_body and header_or_body.
      --help                display this message and exit

Return:
   - 0 in case of normal termination;
   - a non-zero exit status if an error is encountered.""" \
  % os.path.basename(sys.argv[0])


# The only exception defined by this module
class TDBSFCmdLineDecodingError(TDBSFGenericException):
    ExceptionPrettyIdentifier = "Command-line decoding error"


def build_translate_func(where):
    # Return a function converting match spans (i.e. (start, end) tuples) to
    # absolute positions in the file.
    if where == "header":
        def translate_func(match_span, record_info, lhbs):
            shift =  record_info[1]
            return (match_span[0] + shift, match_span[1] + shift)
    elif where == "body":
        def translate_func(match_span, record_info, lhbs):
            shift =  record_info[3]
            return (match_span[0] + shift, match_span[1] + shift)
    else:                               # `where' must be "header_plus_body"
        def translate_func(match_span, record_info, lhbs):
            def translate_start_offset(offset, record_info, lhbs):
                abs_offset = offset + record_info[1]
                if abs_offset <= record_info[2]:
                    return abs_offset
                elif abs_offset < record_info[2] + lhbs:
                    return record_info[2]
                else:
                    return abs_offset + \
                           (record_info[3] - record_info[2]) - lhbs
            def translate_end_offset(offset, record_info, lhbs):
                abs_offset = offset + record_info[1]
                if abs_offset <= record_info[2]:
                    return abs_offset
                elif abs_offset <= record_info[2] + lhbs:
                    return record_info[3]
                else:
                    return abs_offset + \
                           (record_info[3] - record_info[2]) - lhbs
                
            return (
                translate_start_offset(match_span[0], record_info, lhbs),
                translate_end_offset(match_span[1], record_info, lhbs))

    return translate_func


def convert_match_spans_to_absolute(record_info, match_spans, lhbs,
                                    translate_func):
    if type(match_spans) == types.ListType:
        for i in range(len(match_spans[1])):
            match_spans[1][i] = convert_match_spans_to_absolute(
                record_info, match_spans[1][i], lhbs, translate_func)
        return match_spans
    elif type(match_spans) == types.TupleType:
        return translate_func(match_spans, record_info, lhbs)
    else:
        return match_spans              # Should be a single 1


def flatten_match_spans_recurs(match_spans, res):
    if type(match_spans) == types.ListType:
        if match_spans[0] == '~':
            res.append(match_spans)
        else:                           # match_spans[0] must be '&'
           for i in range(len(match_spans[1])):
               index = 0
               elt = match_spans[1][index]
               if type(elt) == types.ListType:
                   flatten_match_spans_recurs(elt, res)
                   del match_spans[1][index]
               elif elt == 1:
                   del match_spans[1][index]
               else:
                   index = index + 1
           if match_spans[1]:           # If there are remaining elements
               res.append(match_spans)
    elif type(match_spans) == types.TupleType:
        res.append(['&', [match_spans]]) # "fake &" (one operand only)
    # match_spans could also be a single 1, in case we add nothing to `res'.


def flatten_match_spans(abs_match_spans):
    # Returns a list containing only ['&', [tuple, tuple, tuple...]] and
    # ['~', [tuple, tuple, tuple...]] elements.
    res = []
    flatten_match_spans_recurs(abs_match_spans, res)
    return res


def emacs_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                    search_string, header_body_separator, db_files_mode,
                    consider_eol_as_lf, global_offset):
    output = sys.stdout

    try:
        search_res = tdbsf_api.lookup_expr(expr, pattern, boh, boh_flags,
                                           eoh, eoh_flags,
                                           search_string,
                                           header_body_separator,
                                           db_files_mode,
                                           consider_eol_as_lf,
                                           global_offset)
    except (tdbsf_api.expr2cpf.TDBSFSyntaxError,
            tdbsf_api.TDBSFInvalidDBFileFormatError,
            tdbsf_api.TDBSFInvalidWhereArg,
            tdbsf_api.TDBSFInvalidNearArg,
            tdbsf_api.cpf2tree.TDBSFRegexpCompilationError), exc_instance:
        output.write("TDBSF error:\n\n%s" % exc_instance.complete_message())
    except IOError:
        output.write("TDBSF error:\n\nI/O error.")
    except:
        output.write("TDBSF error:\n\n\
TDBSF unhandled exception! Please remember precisely how you \
triggered it and report it to Florent!")
    else:
        output.write("(%d . [" % search_res[0]) # number of matches
        for file_res in search_res[1]:
            # Absolute pathname for the current file
            output.write('("%s" . [[' % string.replace(
                # Quote all backslashes for ELisp; it is better than
                # translating them into slashes, I think, because it still
                # allows backslashes in Unix file paths.
                file_res[0], "\\","\\\\"))

            # First, output information about all records contained in the
            # file (not only the records that matched the search expression):
            # boh start, boh end, eoh start and eoh end.
            all_records_info = []
            for record in file_res[1]:
                # We add 1 to every offset in the file contents because buffer
                # positions start at 1 in Emacs.
                all_records_info.append("[(%u . %u) (%u . %u)]"
                                        % (record[0][0] + 1,
                                           record[0][1] + 1,
                                           record[1][0] + 1,
                                           record[1][1] + 1))
            # Write and terminate the array, then start the array containing
            # information about matching records.
            output.write(string.join(all_records_info, ' ') + "]\n[")
            
            info_about_matches = []
            for match in file_res[2]:
                # Convert match spans to absolute offsets in the file and make
                # them Emacs-friendly (meaning the first byte in the file is
                # at offset 1).
                translate_func = build_translate_func(match[1][0])
                abs_match_spans = convert_match_spans_to_absolute(
                    match[0], match[1][1], len(header_body_separator),
                    translate_func)
                # Flatten (and thus degrade, alas) fontification information
                flat_match_spans = flatten_match_spans(abs_match_spans)

                # Turn fontification information into an ELisp structure
                match_info = []
                for group in flat_match_spans:
                    match_spans_list = []
                    for match_span in group[1]:
                        match_spans_list.append("(%s . %s)"
                                                % tuple(map(str, match_span)))
                    
                    match_info.append(
                        '("%s" . [%s])' % (group[0],
                                           string.join(match_spans_list,
                                                       ' ')))
                
                # Turn record-match information into an ELisp structure
                info_about_matches.append("([%s] . [%s])"
                                 % (
                    # Record information
                    string.join(map(str, match[0]), ' '),
                    # Fontification (match) information
                    string.join(match_info, ' ')))

            output.write(string.join(info_about_matches, ' ') + "]])\n")
        output.write("])")

    # Useful for debugging...
#     print ("\nexpr: %s\npattern: %s\nboh: %s\neoh: %s\nsearch_string: %s\n"
#            + "header_body_separator: %s") % (expr, pattern, boh, eoh,
#                                              search_string,
#                                              header_body_separator)
#     print "len(boh): %d\nlen(eoh): %s" % (len(boh), len(eoh))
    
    output.write("\n\n*** End of TDBSF Emacs-oriented output ***\n")
    

def dumb_text_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                        search_string, header_body_separator, db_files_mode,
                        consider_eol_as_lf, global_offset):
    # We do not handle exceptions in this dumb interface to keep it as simple
    # as possible.
    search_res = tdbsf_api.lookup_expr(expr, pattern, boh, boh_flags,
                                       eoh, eoh_flags,
                                       search_string,
                                       header_body_separator,
                                       db_files_mode,
                                       consider_eol_as_lf,
                                       global_offset)
    for file_res in search_res[1]:
        print "%s -> %d record(s) matched:\n" % (file_res[0], len(file_res[2]))
        for match in file_res[2]:
            print "  Record information: %s" % str(match[0])
            print "  Fontification information: %s\n" % str(match[1])
    print "\nTotal: %d matches" % search_res[0]


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "em:s:",
                                   ["boh-flags=",
                                    "crlf-to-lf-in-args",
                                    "db-files-mode=",
                                    "dont-consider-eol-as-lf",
                                    "emacs-friendly-output",
                                    "eoh-flags=",
                                    "help",
                                    "hbs=",
                                    "header-body-separator=",
                                    "search-string="])
    except getopt.GetoptError, message:
        sys.stderr.write(usage + '\n')
        sys.exit(2)

    # Mandatory arguments
    if len(args) != 4:
        sys.stderr.write(usage + '\n')
        sys.exit(2)
    else:
        expr, pattern, boh, eoh = args

    # Default values for options
    boh_flags = ""
    eoh_flags = ""
    # Database files are opened in binary mode by default
    db_files_mode = 'rb'
    emacs_friendly_output = 0
    search_string = "header_or_body"
    header_body_separator = '\xF1'      # LATIN SMALL LETTER N WITH TILDE
    consider_eol_as_lf = 1
    crlf_to_lf_in_args = 0

    for option, value in opts:
        if option == "--boh-flags":
            boh_flags = value
        elif option == "--dont-consider-eol-as-lf":
            consider_eol_as_lf = 0
        elif option == "--crlf-to-lf-in-args":
            crlf_to_lf_in_args = 1
        elif option == "--eoh-flags":
            eoh_flags = value
        elif option in ("-m", "--db-files-mode"):
            if value in ("b", "binary"):
                db_files_mode = "rb"
            elif value in ("t", "text"):
                db_files_mode = "r"
            else:
                sys.stderr.write(usage + '\n')
                sys.exit(2)
        elif option in ("-e", "--emacs-friendly-output"):
            emacs_friendly_output = 1
        elif option == "--help":
            print usage
            sys.exit(0)
        elif option in ("-s", "--search-string"):
            search_string = value
        elif option in ("--hbs", "--header-body-separator"):
            header_body_separator = value
        else:
            raise TDBSFCmdLineDecodingError(
                "unexpected option received from the getopt module; \
probably a bug in %s" % sys.argv[0])

    if crlf_to_lf_in_args:
        expr = string.replace(expr, "\r\n", "\n")
        boh = string.replace(boh, "\r\n", "\n")
        eoh = string.replace(eoh, "\r\n", "\n")
        header_body_separator = string.replace(header_body_separator,
                                               "\r\n", "\n")

    if emacs_friendly_output:
        # Within Emacs, buffer positions start at 1, not 0, hence the last
        # argument's value to offset all positions specifying the beginning of
        # the "boh" part of a record, the end of this part, the beginning of
        # its "eoh" part, its end, and the end of the record's body.
        emacs_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                        search_string, header_body_separator, db_files_mode,
                        consider_eol_as_lf, 1)
    else:
        # In this dumb interface, we let the file positions start at 0.
        dumb_text_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                            search_string, header_body_separator,
                            db_files_mode, consider_eol_as_lf, 0)

    sys.exit(0)


if __name__ == "__main__": main()
