;;; tdbsf-user-interface.el --- TDBSF ELisp end-user interface

;; Copyright (c) 2001 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(require 'tdbsf)

(defvar tdbsf-default-pattern
  "/home/flo/src/tdbsf/db_papa/*.asc"
  "Default file globbing pattern to use by the TDBSF engine.
This pattern describes (in the syntax of the Python 'glob' module)
the set of files that should be examined by the TDBSF engine.

The syntax used by the 'glob' module should be relatively close to
that of POSIX shell file globbing.")

(defvar tdbsf-default-search-string
  "header or body"
  "Default string to use by TDBSF to evaluate the search expression.
The TDBSF engine uses a search expression to decide whether a record
does or does not \"match\". This search expression can be evaluated
on the header part of each record, on its body part or on a string
that is the result of the concatenation of the header, a custom
separator string and the body. It can also consider that a record
matches if (and only if) either the header string or the body string
matches the search expression.

The `tdbsf' command allows the user to select the desired behavior
between these possibilities. This variable controls the default
setting when that question is asked. It can be set to one of the
following values (strings):
  - \"header\"
  - \"body\"
  - \"header + custom separation string + body (all concatenated)\"
  - \"header or body\"

which map respectively to the options described above.")

(defvar tdbsf-default-header-body-separator
  ""
  "Default separator to use by TDBSF between a record's header and its body.
The TDBSF engine uses a search expression to decide whether a record
does or does not \"match\". This search expression can be evaluated
on the header part of each record, on its body part or on a string
that is the result of the concatenation of the header, a custom
separator string and the body. It can also consider that a record
matches if (and only if) either the header string or the body string
matches the search expression.

The `tdbsf' command allows the user to select the desired behavior
between these possibilities. When the third option is chosen (i.e.,
the string used to evaluate the search expression is header +
separator + body), this command asks the user which string he wants
to use as the separator. This variable controls the default
setting when that question is asked. It can be set to any string.")

(defvar tdbsf-boh-pattern
  "^ *{60,} *\n"
  "A pattern identifying \"Trouve\"'s ugly beginning of header string.")

(defvar tdbsf-boh-flags
  "MULTILINE"
  "The flags going with `tdbsf-boh-pattern' to define a regular expression \
for the \"re\" Python module.")

(defvar tdbsf-eoh-pattern
  "^ *{60,} *\n"
  "A pattern identifying \"Trouve\"'s ugly end of header string.")

(defvar tdbsf-eoh-flags
  "MULTILINE"
  "The flags going with `tdbsf-eoh-pattern' to define a regular expression \
for the \"re\" Python module.")

(defvar tdbsf-search-string-alist
  '(("header" . "header")
    ("body" . "body")
    ("header or body" . "header_or_body")
    ("header + custom separation string + body (all concatenated)"
     . "header_plus_body"))
  "Alist mapping displayed values to arguments for --search-string in TDBSF.
Alist specifying the mapping between the displayed values when the user
is asked by the `tdbsf' command to the arguments that are passed to
the TDBSF engine for the various possible values of --search-string.")

(defvar tdbsf-db-files-mode
  "binary"
  "Mode to open database files in.

Mode to use by the TDBSF engine when opening TDBSF database files.
Permitted values are \"binary\" (or \"b\") and \"text\" (or \"t\").")

(defvar tdbsf-crlf-to-lf-in-args
  t
  "Whether the TDBSF engine should convert CRLF to LF in arguments.

CR = Carriage Return = ASCII character number 13 (in decimal)
LF = Line Feed       = ASCII character number 10 (in decimal)")

(defvar tdbsf-dont-consider-eol-as-lf
  nil
  "Don't try to guess the end of line type in TDBSF database files.
If t, the TDBSF engine will work on the raw contents of the database
files. If nil, it will first try to guess the end of line type of
each database file and convert it to LF (Line Feed, i.e. ASCII
character 10 in decimal).")

(defun tdbsf (expr pattern search-string header-body-separator)
  "Gather the required information and call the TDBSF engine.
Ask some information about the search to perform, pass this
information to the TDBSF engine and report the results in a summary
buffer.

You can customize the following variables to change the default
answers to the questions asked by this command:
  - tdbsf-default-pattern
  - tdbsf-default-search-string (and, less importantly,
    tdbsf-search-string-alist)
  - tdbsf-default-header-body-separator

Additionally, tdbsf-boh-pattern, tdbsf-eoh-pattern, tdbsf-boh-flags,
tdbsf-eoh-flags, tdbsf-db-files-mode, tdbsf-crlf-to-lf-in-args and
tdbsf-dont-consider-eol-as-lf can be used to further control the
behavior of the TDBSF engine."
  (interactive
   (let*
       (;; Pattern
        (pattern
         (let ((insert-default-directory nil))
           (expand-file-name (read-file-name "Pattern: " nil nil nil
                                             tdbsf-default-pattern))))
        ;; Search string ("header", "header_plus_body", etc.)
        (search-string
         (cdr (assoc (completing-read
                      (concat
                       "What part of a record should \"match\": (default "
                       tdbsf-default-search-string ") ")
                      tdbsf-search-string-alist
                      nil t nil nil tdbsf-default-search-string)
                     tdbsf-search-string-alist)))
        ;; Separator string to insert in the "header_plus_body" case
        (header-body-separator
         (if (string-equal search-string "header_plus_body")
         (read-string
          (concat 
           "Separator string to insert between header and body: (default "
           tdbsf-default-header-body-separator ") ")
          nil nil tdbsf-default-header-body-separator)
         tdbsf-default-header-body-separator ; does not matter
         )))
     (list (buffer-substring-no-properties (region-beginning) (region-end))
           pattern search-string header-body-separator)))
  (call-tdbsf expr pattern tdbsf-boh-pattern tdbsf-boh-flags
              tdbsf-eoh-pattern tdbsf-eoh-flags
              search-string header-body-separator tdbsf-db-files-mode
              tdbsf-crlf-to-lf-in-args tdbsf-dont-consider-eol-as-lf))

(provide 'tdbsf-user-interface)
